"""
.. currentmodule:: flytekitplugins.polars

This package contains things that are useful when extending Flytekit.

.. autosummary::
   :template: custom.rst
   :toctree: generated/

   PolarsDataFrameToParquetEncodingHandler
   ParquetToPolarsDataFrameDecodingHandler
"""

from .sd_transformers import ParquetToPolarsDataFrameDecodingHandler, PolarsDataFrameToParquetEncodingHandler



import io
import typing

from flytekit import FlyteContext, lazy_module
from flytekit.core.data_persistence import get_fsspec_storage_options
from flytekit.models import literals
from flytekit.models.literals import StructuredDatasetMetadata
from flytekit.models.types import StructuredDatasetType
from flytekit.types.structured.structured_dataset import (
    PARQUET,
    StructuredDataset,
    StructuredDatasetDecoder,
    StructuredDatasetEncoder,
    StructuredDatasetTransformerEngine,
)

if typing.TYPE_CHECKING:
    import fsspec.utils as fsspec_utils

    import polars as pl
else:
    pl = lazy_module("polars")
    fsspec_utils = lazy_module("fsspec.utils")


############################
# Polars DataFrame classes #
############################


class PolarsDataFrameRenderer:
    """
    The Polars DataFrame summary statistics are rendered as an HTML table.
    """

    def to_html(self, df: typing.Union[pl.DataFrame, pl.LazyFrame]) -> str:
        assert isinstance(df, (pl.DataFrame, pl.LazyFrame))
        try:
            describe_df = df.describe()
        except AttributeError:
            # LazyFrames in polars <= 0.19 does not support `describe`
            describe_df = df.collect().describe()

        # the value is "statistic" or "describe" depending on polars version
        stat_colname = describe_df.columns[0]

        columns = describe_df[stat_colname]
        html_repr = describe_df.drop(stat_colname).transpose(column_names=columns)._repr_html_()
        return html_repr


class PolarsDataFrameToParquetEncodingHandler(StructuredDatasetEncoder):
    def __init__(self):
        super().__init__(pl.DataFrame, None, PARQUET)

    def encode(
        self,
        ctx: FlyteContext,
        structured_dataset: StructuredDataset,
        structured_dataset_type: StructuredDatasetType,
    ) -> literals.StructuredDataset:
        df = typing.cast(pl.DataFrame, structured_dataset.dataframe)

        output_bytes = io.BytesIO()
        # Polars 0.13.12 deprecated to_parquet in favor of write_parquet
        if hasattr(df, "write_parquet"):
            df.write_parquet(output_bytes)
        else:
            df.to_parquet(output_bytes)

        if structured_dataset.uri is not None:
            fs = ctx.file_access.get_filesystem_for_path(path=structured_dataset.uri)
            with fs.open(structured_dataset.uri, "wb") as s:
                s.write(output_bytes)
            output_uri = structured_dataset.uri
        else:
            remote_fn = "00000"  # 00000 is our default unnamed parquet filename
            output_uri = ctx.file_access.put_raw_data(output_bytes, file_name=remote_fn)
        return literals.StructuredDataset(uri=output_uri, metadata=StructuredDatasetMetadata(structured_dataset_type))


class ParquetToPolarsDataFrameDecodingHandler(StructuredDatasetDecoder):
    def __init__(self):
        super().__init__(pl.DataFrame, None, PARQUET)

    def decode(
        self,
        ctx: FlyteContext,
        flyte_value: literals.StructuredDataset,
        current_task_metadata: StructuredDatasetMetadata,
    ) -> pl.DataFrame:
        uri = flyte_value.uri

        kwargs = get_fsspec_storage_options(
            protocol=fsspec_utils.get_protocol(uri),
            data_config=ctx.file_access.data_config,
        )
        if current_task_metadata.structured_dataset_type and current_task_metadata.structured_dataset_type.columns:
            columns = [c.name for c in current_task_metadata.structured_dataset_type.columns]
            return pl.read_parquet(uri, columns=columns, use_pyarrow=True, storage_options=kwargs)
        return pl.read_parquet(uri, use_pyarrow=True, storage_options=kwargs)


############################
# Polars LazyFrame classes #
############################


class PolarsLazyFrameToParquetEncodingHandler(StructuredDatasetEncoder):
    def __init__(self):
        super().__init__(pl.LazyFrame, None, PARQUET)

    def encode(
        self,
        ctx: FlyteContext,
        structured_dataset: StructuredDataset,
        structured_dataset_type: StructuredDatasetType,
    ) -> literals.StructuredDataset:
        lf = typing.cast(pl.LazyFrame, structured_dataset.dataframe)

        # The pl.LazyFrame.sink_parquet method uses streaming mode, which is currently considered unstable. Until it is
        # stable, we collect the dataframe and write it to a BytesIO buffer.
        df = lf.collect()

        if hasattr(df, "write_parquet"):
            # Polars 0.13.12 deprecated to_parquet in favor of write_parquet
            _write_method = df.write_parquet
        else:
            _write_method = df.to_parquet

        if structured_dataset.uri is not None:
            fs = ctx.file_access.get_filesystem_for_path(path=structured_dataset.uri)
            with fs.open(structured_dataset.uri, "wb") as s:
                _write_method(s)
            output_uri = structured_dataset.uri
        else:
            output_bytes = io.BytesIO()
            remote_fn = "00000"  # 00000 is our default unnamed parquet filename
            _write_method(output_bytes)
            output_uri = ctx.file_access.put_raw_data(output_bytes, file_name=remote_fn)
        return literals.StructuredDataset(uri=output_uri, metadata=StructuredDatasetMetadata(structured_dataset_type))


class ParquetToPolarsLazyFrameDecodingHandler(StructuredDatasetDecoder):
    def __init__(self):
        super().__init__(pl.LazyFrame, None, PARQUET)

    def decode(
        self,
        ctx: FlyteContext,
        flyte_value: literals.StructuredDataset,
        current_task_metadata: StructuredDatasetMetadata,
    ) -> pl.LazyFrame:
        uri = flyte_value.uri

        kwargs = get_fsspec_storage_options(
            protocol=fsspec_utils.get_protocol(uri),
            data_config=ctx.file_access.data_config,
        )
        # use read_parquet instead of scan_parquet for now because scan_parquet currently doesn't work with fsspec:
        # https://github.com/pola-rs/polars/issues/16737
        if current_task_metadata.structured_dataset_type and current_task_metadata.structured_dataset_type.columns:
            columns = [c.name for c in current_task_metadata.structured_dataset_type.columns]
            return pl.read_parquet(uri, columns=columns, use_pyarrow=True, storage_options=kwargs).lazy()
        return pl.read_parquet(uri, use_pyarrow=True, storage_options=kwargs).lazy()


# Register the Polars DataFrame handlers
StructuredDatasetTransformerEngine.register(PolarsDataFrameToParquetEncodingHandler())
StructuredDatasetTransformerEngine.register(ParquetToPolarsDataFrameDecodingHandler())
StructuredDatasetTransformerEngine.register_renderer(pl.DataFrame, PolarsDataFrameRenderer())

# Register the Polars LazyFrame handlers
StructuredDatasetTransformerEngine.register(PolarsLazyFrameToParquetEncodingHandler())
StructuredDatasetTransformerEngine.register(ParquetToPolarsLazyFrameDecodingHandler())



import tempfile

import pandas as pd
import polars as pl
import pytest
from flytekitplugins.polars.sd_transformers import PolarsDataFrameRenderer
from typing_extensions import Annotated
from packaging import version
from polars.testing import assert_frame_equal

from flytekit import kwtypes, task, workflow
from flytekit.types.structured.structured_dataset import PARQUET, StructuredDataset

subset_schema = Annotated[StructuredDataset, kwtypes(col2=str), PARQUET]
full_schema = Annotated[StructuredDataset, PARQUET]

polars_version = pl.__version__


@pytest.mark.parametrize("df_cls", [pl.DataFrame, pl.LazyFrame])
def test_polars_workflow_subset(df_cls):
    @task
    def generate() -> subset_schema:
        df = df_cls({"col1": [1, 3, 2], "col2": list("abc")})
        return StructuredDataset(dataframe=df)

    @task
    def consume(df: subset_schema) -> subset_schema:
        df = df.open(df_cls).all()

        materialized_df = df
        if df_cls is pl.LazyFrame:
            materialized_df = df.collect()

        assert materialized_df["col2"][0] == "a"
        assert materialized_df["col2"][1] == "b"
        assert materialized_df["col2"][2] == "c"

        return StructuredDataset(dataframe=df)

    @workflow
    def wf() -> subset_schema:
        return consume(df=generate())

    result = wf()
    assert result is not None


@pytest.mark.parametrize("df_cls", [pl.DataFrame, pl.LazyFrame])
def test_polars_workflow_full(df_cls):
    @task
    def generate() -> full_schema:
        df = df_cls({"col1": [1, 3, 2], "col2": list("abc")})
        return StructuredDataset(dataframe=df)

    @task
    def consume(df: full_schema) -> full_schema:
        df = df.open(df_cls).all()

        materialized_df = df
        if df_cls is pl.LazyFrame:
            materialized_df = df.collect()

        assert materialized_df["col1"][0] == 1
        assert materialized_df["col1"][1] == 3
        assert materialized_df["col1"][2] == 2
        assert materialized_df["col2"][0] == "a"
        assert materialized_df["col2"][1] == "b"
        assert materialized_df["col2"][2] == "c"

        return StructuredDataset(dataframe=df.sort("col1"))

    @workflow
    def wf() -> full_schema:
        return consume(df=generate())

    result = wf()
    assert result is not None


@pytest.mark.parametrize("df_cls", [pl.DataFrame, pl.LazyFrame])
def test_polars_renderer(df_cls):
    df = df_cls({"col1": [1, 3, 2], "col2": list("abc")})

    if df_cls is pl.LazyFrame:
        df_desc = df.collect().describe()
    else:
        df_desc = df.describe()

    stat_colname = df_desc.columns[0]
    expected = df_desc.drop(stat_colname).transpose(column_names=df_desc[stat_colname])._repr_html_()
    assert PolarsDataFrameRenderer().to_html(df) == expected


@pytest.mark.parametrize("df_cls", [pl.DataFrame, pl.LazyFrame])
def test_parquet_to_polars_dataframe(df_cls):
    data = {"name": ["Alice"], "age": [5]}

    @task
    def create_sd() -> StructuredDataset:
        df = df_cls(data=data)
        return StructuredDataset(dataframe=df)

    sd = create_sd()
    polars_df = sd.open(df_cls).all()
    if isinstance(polars_df, pl.LazyFrame):
        polars_df = polars_df.collect()

    assert_frame_equal(pl.DataFrame(data), polars_df)

    tmp = tempfile.mktemp()
    pl.DataFrame(data).write_parquet(tmp)

    @task
    def consume_sd_return_df(sd: StructuredDataset) -> df_cls:
        return sd.open(df_cls).all()

    sd = StructuredDataset(uri=tmp)
    df_out = consume_sd_return_df(sd=sd)

    if df_cls is pl.LazyFrame:
        df_out = df_out.collect()

    assert_frame_equal(df_out, polars_df)

    @task
    def consume_sd_return_sd(sd: StructuredDataset) -> StructuredDataset:
        return StructuredDataset(dataframe=sd.open(df_cls).all())

    sd = StructuredDataset(uri=tmp)
    opened_sd = consume_sd_return_sd(sd=sd).open(df_cls).all()

    if df_cls is pl.LazyFrame:
        opened_sd = opened_sd.collect()

    assert_frame_equal(opened_sd, polars_df)
